from lxml import etree as ET
import io
from collections import namedtuple


IXMLTrack = namedtuple('IXMLTrack', ['channel_index', 'interleave_index', 'name', 'function'])


class WavIXMLFormat:
    """
    iXML recorder metadata.
    """
    def __init__(self, xml):
        """
        Parse iXML.
        :param xml: A bytes-like object containing the iXML payload.
        """
        self.source = xml
        xml_bytes = io.BytesIO(xml)
        parser = ET.XMLParser(recover=True)
        self.parsed = ET.parse(xml_bytes, parser=parser)

    def _get_text_value(self, xpath):
        e = self.parsed.find("./" + xpath)
        if e is not None:
            return e.text

    @property
    def raw_xml(self):
        """
        The root entity of the iXML document.
        """
        return self.parsed

    @property
    def track_list(self):
        """
        A description of each track.
        :return: An Iterator
        """
        for track in self.parsed.find("./TRACK_LIST").iter():
            if track.tag == 'TRACK':
                yield IXMLTrack(channel_index=track.xpath('string(CHANNEL_INDEX/text())'),
                                interleave_index=track.xpath('string(INTERLEAVE_INDEX/text())'),
                                name=track.xpath('string(NAME/text())'),
                                function=track.xpath('string(FUNCTION/text())'))

    @property
    def project(self):
        """
        The project/film name entered for the recording.
        """
        return self._get_text_value("PROJECT")

    @property
    def scene(self):
        """
        Scene/slate.
        """
        return self._get_text_value("SCENE")

    @property
    def take(self):
        """
        Take number.
        """
        return self._get_text_value("TAKE")

    @property
    def tape(self):
        """
        Tape name.
        """
        return self._get_text_value("TAPE")

    @property
    def family_uid(self):
        """
        The globally-unique ID for this file family. This may be in the format
        of a GUID, or an EBU Rec 9 source identifier, or some other dumb number.
        """
        return self._get_text_value("FILE_SET/FAMILY_UID")

    @property
    def family_name(self):
        """
        The name of this file's file family.
        """
        return self._get_text_value("FILE_SET/FAMILY_NAME")



# Dolby RMU Metadata per EBU Tech 3285 Supp 6
#
# https://tech.ebu.ch/docs/tech/tech3285s6.pdf
#

from struct import unpack, calcsize
from enum import (Enum, IntEnum)

CHUNK_IDENT = "dbmd"
DOLBY_VERSION = "1.0.0.6"


class _DPPGenericDownMixLevel(Enum):
    PLUS_3DB = 0b000
    PLUS_1_5DB = 0b001
    UNITY = 0b010
    MINUS_1_5DB = 0b011
    MINUS_3DB = 0b100
    MINUS_4_5DB = 0b101
    MINUS_6DB = 0b110
    MUTE = 0b111


class DPPDolbySurroundEncodingMode(Enum):
    RESERVED = 0b11
    IN_USE = 0b10
    NOT_IN_USE = 0b01
    NOT_INDICATED = 0b00

# DPPLoRoDownMixCenterLevel
# DPPLtRtCenterMixLevel
# DPPLtRtSurroundMixLevel

class DolbyMetadataSegmentTypes(IntEnum):
    END_MARKER = 0
    DOLBY_E_METADATA = 1
    DOLBY_DIGITAL_METADATA = 3
    DOLBY_DIGITAL_PLUS_METADATA = 7
    AUDIO_INFO = 8


class DDPBitStreamMode(Enum):
    """
    Dolby Digital Plus `bsmod` field
    § 4.3.2.2
    """
    COMPLETE_MAIN = 0b000
    MUSIC_AND_EFFECTS = 0b001
    VISUALLY_IMPAIRED = 0b010
    HEARING_IMPAIRED = 0b011
    DIALOGUE_ONLY = 0b100
    COMMENTARY = 0b101
    EMERGENCY = 0b110
    VOICEOVER = 0b111 # if audioconfigmode is 1_0
    KARAOKE = 0b1000  # if audioconfigmode is not 1_0


class DDPAudioCodingMode(Enum):
    """
    Dolby Digital Plus `acmod` field
    § 4.3.2.3
    """
    RESERVED = 0b000
    CH_ORD_1_0 = 0b001
    CH_ORD_2_0 = 0b010
    CH_ORD_3_0 = 0b011
    CH_ORD_2_1 = 0b100
    CH_ORD_3_1 = 0b101
    CH_ORD_2_2 = 0b110
    CH_ORD_3_2 = 0b111


class DPPCenterDownMixLevel(Enum):
    """
    § 4.3.3.1
    """
    DOWN_3DB = 0b00
    DOWN_45DB = 0b01
    DOWN_6DB = 0b10
    RESERVED = 0b11


class DPPSurroundDownMixLevel(Enum):
    """
    Dolby Digital Plus `surmixlev` field
    § 4.3.3.2
    """
    DOWN_3DB = 0b00
    DOWN_6DB = 0b01
    MUTE = 0b10
    RESERVED = 0b11


class DPPLanguageCode(Enum):
    """
    § 4.3.4.1 , 4.3.5 (always 0xFF)
    """
    # this is removed in https://www.atsc.org/wp-content/uploads/2015/03/A52-201212-17.pdf § 5.4.2.12
    # It should just be 0xff
    pass


class DPPMixLevel(int):
    pass


class DPPDialnormLevel(int):
    pass


class DPPRoomTime(Enum):
    """
    `roomtyp` 4.3.6.3
    """
    NOT_INDICATED = 0b00
    LARGE_ROOM_X_CURVE = 0b01
    SMALL_ROOM_FLAT_CURVE = 0b10
    RESERVED = 0b11


class DPPPreferredDownMixMode(Enum):
    """
    § 4.3.8.1
    """
    NOT_INDICATED = 0b00
    PRO_LOGIC = 0b01
    STEREO = 0b10
    PRO_LOGIC_2 = 0b11


# class DPPLtRtCenterMixLevel(_DPPGenericDownMixLevel):
#     pass
#
#
# class DPPLtRtSurroundMixLevel(_DPPGenericDownMixLevel):
#     pass
#
#
# class DPPSurroundEXMode(_DPPGenericInUseIndicator):
#     pass
#
#
# class DPPHeadphoneMode(_DPPGenericInUseIndicator):
#     pass


class DPPADConverterType(Enum):
    STANDARD = 0
    HDCD = 1


class DDPStreamDependency(Enum):
    """
    Encodes `ddplus_info1.stream_type` field § 4.3.12.1
    """
    INDEPENDENT = 0
    DEPENDENT = 1
    INDEPENDENT_FROM_DOLBY_DIGITAL = 2
    RESERVED = 3


class DDPDataRate(int):
    pass


class DPPRFCompressionProfile(Enum):
    NONE = 0
    FILM_STANDARD = 1
    FILM_LIGHT = 2
    MUSIC_STANDARD = 3
    MUSIC_LIGHT = 4
    SPEECH = 5


class DolbyDigitalPlusMetadata:

    @classmethod
    def parse(cls, binary_data):
        binary_format = "<BBxxBBBBBBBBBBBBBxxxBxxxxxH"
        assert len(binary_data >= calcsize(binary_format))
        fields = unpack(binary_format, binary_data)


class WavDolbyReader:
    def __init__(self, dolby_data):
        version, remainder = unpack("<U", dolby_data[0]), dolby_data[1:]

        ## FIXME continues...


from optparse import OptionParser, OptionGroup
import datetime
from . import WavInfoReader
import sys
import json


def main():
    parser = OptionParser()

    parser.usage = 'wavinfo [FILE.wav]*'

    # parser.add_option('-f', dest='output_format', help='Set the output format',
    #                   default='json',
    #                   metavar='FORMAT')

    (options, args) = parser.parse_args(sys.argv)
    for arg in args[1:]:
        try:
            this_file = WavInfoReader(path=arg)
            ret_dict = {'file_argument': arg, 'run_date': datetime.datetime.now().isoformat() , 'scopes': {}}
            for scope, name, value in this_file.walk():
                if scope not in ret_dict['scopes'].keys():
                    ret_dict['scopes'][scope] = {}

                ret_dict['scopes'][scope][name] = value

            json.dump(ret_dict, fp=sys.stdout, indent=2)
        except Exception as e:
            print(e)


if __name__ == "__main__":
    main()




import struct
from collections import namedtuple
from .rf64_parser import parse_rf64


class WavInfoEOFError(EOFError):
    def __init__(self, identifier, chunk_start):
        self.identifier = identifier
        self.chunk_start = chunk_start


class ListChunkDescriptor(namedtuple('ListChunkDescriptor', 'signature children')):
    pass
    # def find(self, chunk_path):
    #     if len(chunk_path) > 1:
    #         for chunk in self.children:
    #             if type(chunk) is ListChunkDescriptor and \
    #                     chunk.signature is chunk_path[0]:
    #                 return chunk.find(chunk_path[1:])
    #     else:
    #         for chunk in self.children:
    #             if type(chunk) is ChunkDescriptor and \
    #                     chunk.ident is chunk_path[0]:
    #                 return chunk


class ChunkDescriptor(namedtuple('ChunkDescriptor', 'ident start length rf64_context')):
    def read_data(self, from_stream):
        from_stream.seek(self.start)
        return from_stream.read(self.length)


def parse_list_chunk(stream, length, rf64_context=None):
    start = stream.tell()
    signature = stream.read(4)

    children = []
    while stream.tell() - start + 8 < length:
        child_chunk = parse_chunk(stream, rf64_context=rf64_context)
        children.append(child_chunk)

    stream.seek(start + length)

    return ListChunkDescriptor(signature=signature, children=children)


def parse_chunk(stream, rf64_context=None):
    header_start = stream.tell()
    ident = stream.read(4)
    size_bytes = stream.read(4)

    if len(ident) != 4 or len(size_bytes) != 4:
        raise WavInfoEOFError(identifier=ident, chunk_start=header_start)

    data_size = struct.unpack('<I', size_bytes)[0]

    if data_size == 0xFFFFFFFF:
        if rf64_context is None and ident in {b'RF64', b'BW64'}:
            rf64_context = parse_rf64(stream=stream, signature=ident)

        data_size = rf64_context.bigchunk_table[ident]

    displacement = data_size
    if displacement % 2:
        displacement += 1

    if ident in {b'RIFF', b'LIST', b'RF64', b'BW64'}:
        return parse_list_chunk(stream=stream, length=data_size, rf64_context=rf64_context)
    else:
        data_start = stream.tell()
        stream.seek(displacement, 1)
        return ChunkDescriptor(ident=ident, start=data_start, length=data_size, rf64_context=rf64_context)



"""
methods to probe a WAV file for various kinds of production metadata.

Go to the documentation for wavinfo.WavInfoReader for more information.
"""

from .wave_reader import WavInfoReader
from .riff_parser import WavInfoEOFError

__version__ = '1.6.2'
__author__ = 'Jamie Hardt <jamiehardt@gmail.com>'
__license__ = "MIT"



import struct
from collections import namedtuple
from . import riff_parser

RF64Context = namedtuple('RF64Context','sample_count bigchunk_table')


def parse_rf64(stream, signature = b'RF64'):
    # print("starting parse_rf64")
    start = stream.tell()
    assert( stream.read(4) == b'WAVE' )

    ds64_chunk = riff_parser.parse_chunk(stream)

    ds64_field_spec = "<QQQI"
    ds64_fields_size = struct.calcsize(ds64_field_spec)
    assert(ds64_chunk.ident == b'ds64')

    ds64_data = ds64_chunk.read_data(stream)
    assert(len(ds64_data) >= ds64_fields_size )

    # print("Read ds64 chunk: len()",len(ds64_data))
    riff_size, data_size, sample_count, length_lookup_table = struct.unpack( ds64_field_spec , ds64_data[0:ds64_fields_size] )

    bigchunk_table = {}
    chunksize64format = "<4sL"
    chunksize64size = struct.calcsize(chunksize64format)
    # print("Found chunks64s:", length_lookup_table)

    for n in range(length_lookup_table):
        bigname, bigsize = struct.unpack_from( chunksize64format , ds64_data, offset= ds64_fields_size )
        bigchunk_table[bigname] = bigsize

    bigchunk_table[b'data'] = data_size
    bigchunk_table[signature] = riff_size

    stream.seek(start, 0)
    # print("returning from parse_rf64, context: ", RF64Context(sample_count=sample_count, bigchunk_table=bigchunk_table))
    return RF64Context( sample_count=sample_count, bigchunk_table=bigchunk_table )




from .riff_parser import parse_chunk, ListChunkDescriptor


class WavInfoChunkReader:

    def __init__(self, f, encoding):
        self.encoding = encoding

        f.seek(0)
        parsed_chunks = parse_chunk(f)

        list_chunks = [chunk for chunk in parsed_chunks.children if type(chunk) is ListChunkDescriptor]

        self.info_chunk = next((chunk for chunk in list_chunks if chunk.signature == b'INFO'), None)

        #: 'ICOP' Copyright
        self.copyright = self._get_field(f, b'ICOP')
        #: 'IPRD' Product
        self.product = self._get_field(f, b'IPRD')
        self.album = self.product
        #: 'IGNR' Genre
        self.genre = self._get_field(f, b'IGNR')
        #: 'ISBJ' Supject
        self.subject = self._get_field(f, b'ISBJ')
        #: 'IART' Artist, composer, author
        self.artist = self._get_field(f, b'IART')
        #: 'ICMT' Comment
        self.comment = self._get_field(f, b'ICMT')
        #: 'ISFT' Software, encoding application
        self.software = self._get_field(f, b'ISFT')
        #: 'ICRD' Created date
        self.created_date = self._get_field(f, b'ICRD')
        #: 'IENG' Engineer
        self.engineer = self._get_field(f, b'IENG')
        #: 'ITCH' Technician
        self.technician = self._get_field(f, b'ITCH')
        #: 'IKEY' Keywords, keyword list
        self.keywords = self._get_field(f, b'IKEY')
        #: 'INAM' Name, title
        self.title = self._get_field(f, b'INAM')
        #: 'ISRC' Source
        self.source = self._get_field(f, b'ISRC')
        #: 'TAPE' Tape
        self.tape = self._get_field(f, b'TAPE')
        #: 'IARL' Archival Location
        self.archival_location = self._get_field(f, b'IARL')
        #: 'ICSM' Commissioned
        self.commissioned = self._get_field(f, b'ICMS')

    def _get_field(self, f, field_ident):
        search = next(((chunk.start, chunk.length) for chunk in self.info_chunk.children if chunk.ident == field_ident),
                      None)

        if search is not None:
            f.seek(search[0])
            data = f.read(search[1])
            return data.decode(self.encoding).rstrip('\0')
        else:
            return None

    def to_dict(self):
        """
        A dictionary with all of the key/values read from the INFO scope.
        """
        return {'copyright': self.copyright,
                'product': self.product,
                'album': self.album,
                'genre': self.genre,
                'artist': self.artist,
                'comment': self.comment,
                'software': self.software,
                'created_date': self.created_date,
                'engineer': self.engineer,
                'keywords': self.keywords,
                'title': self.title,
                'source': self.source,
                'tape': self.tape,
                'commissioned': self.commissioned,
                'archival_location': self.archival_location,
                'subject': self.subject,
                'technician': self.technician
                }

    def __repr__(self):
        return_val = self.to_dict()
        return_val.update({'encoding': self.encoding})
        return str(return_val)



from typing import Union
import binascii
from functools import reduce


def binary_to_string(binary_value):
    return reduce(lambda val, el: val + "{:02x}".format(el), binary_value, '')


class UMIDParser:
    """
    Parse a raw binary SMPTE 330M Universal Materials Identifier

    This implementation is based on SMPTE ST 330:2011
    """
    def __init__(self, raw_umid: bytearray):
        self.raw_umid = raw_umid
    #
    # @property
    # def universal_label(self) -> bytearray:
    #     return self.raw_umid[0:12]
    #
    # @property
    # def basic_umid(self):
    #     return self.raw_umid[0:32]

    def basic_umid_to_str(self):
        return binary_to_string(self.raw_umid[0:32])
    #
    # @property
    # def universal_label_is_valid(self) -> bool:
    #     valid_preamble = b'\x06\x0a\x2b\x34\x01\x01\x01\x05\x01\x01'
    #     return self.universal_label[0:len(valid_preamble)] == valid_preamble
    #
    # @property
    # def material_type(self) -> str:
    #     material_byte = self.raw_umid[10]
    #     if material_byte == 0x1:
    #         return 'picture'
    #     elif material_byte == 0x2:
    #         return 'audio'
    #     elif material_byte == 0x3:
    #         return 'data'
    #     elif material_byte == 0x4:
    #         return 'other'
    #     elif material_byte == 0x5:
    #         return 'picture_single_component'
    #     elif material_byte == 0x6:
    #         return 'picture_multiple_component'
    #     elif material_byte == 0x7:
    #         return 'audio_single_component'
    #     elif material_byte == 0x9:
    #         return 'audio_multiple_component'
    #     elif material_byte == 0xb:
    #         return 'auxiliary_single_component'
    #     elif material_byte == 0xc:
    #         return 'auxiliary_multiple_component'
    #     elif material_byte == 0xd:
    #         return 'mixed_components'
    #     elif material_byte == 0xf:
    #         return 'not_identified'
    #     else:
    #         return 'not_recognized'
    #
    # @property
    # def material_number_creation_method(self) -> str:
    #     method_byte = self.raw_umid[11]
    #     method_byte = (method_byte << 4) & 0xf
    #     if method_byte == 0x0:
    #         return 'undefined'
    #     elif method_byte == 0x1:
    #         return 'smpte'
    #     elif method_byte == 0x2:
    #         return 'uuid'
    #     elif method_byte == 0x3:
    #         return 'masked'
    #     elif method_byte == 0x4:
    #         return 'ieee1394'
    #     elif 0x5 <= method_byte <= 0x7:
    #         return 'reserved_undefined'
    #     else:
    #         return 'unrecognized'
    #
    # @property
    # def instance_number_creation_method(self) -> str:
    #     method_byte = self.raw_umid[11]
    #     method_byte = method_byte & 0xf
    #     if method_byte == 0x0:
    #         return 'undefined'
    #     elif method_byte == 0x01:
    #         return 'local_registration'
    #     elif method_byte == 0x02:
    #         return '24_bit_prs'
    #     elif method_byte == 0x03:
    #         return 'copy_number_and_16_bit_prs'
    #     elif 0x04 <= method_byte <= 0x0e:
    #         return 'reserved_undefined'
    #     elif method_byte == 0x0f:
    #         return 'live_stream'
    #     else:
    #         return 'unrecognized'
    #
    # @property
    # def indicated_length(self) -> str:
    #     if self.raw_umid[12] == 0x13:
    #         return 'basic'
    #     elif self.raw_umid[12] == 0x33:
    #         return 'extended'
    #
    # @property
    # def instance_number(self) -> bytearray:
    #     return self.raw_umid[13:3]
    #
    # @property
    # def material_number(self) -> bytearray:
    #     return self.raw_umid[16:16]
    #
    # @property
    # def source_pack(self) -> Union[bytearray, None]:
    #     if self.indicated_length == 'extended':
    #         return self.raw_umid[32:32]
    #     else:
    #         return None



#-*- coding: utf-8 -*-
import struct
import os
from collections import namedtuple

import pathlib

from .riff_parser import parse_chunk, ChunkDescriptor, ListChunkDescriptor
from .wave_ixml_reader import WavIXMLFormat
from .wave_bext_reader import WavBextReader
from .wave_info_reader import WavInfoChunkReader

#: Calculated statistics about the audio data.
WavDataDescriptor = namedtuple('WavDataDescriptor', 'byte_count frame_count')

#: The format of the audio samples.
WavAudioFormat = namedtuple('WavAudioFormat',
                            'audio_format channel_count sample_rate byte_rate block_align bits_per_sample')


class WavInfoReader:
    """
    Parse a WAV audio file for metadata.
    """

    def __init__(self, path, info_encoding='latin_1', bext_encoding='ascii'):
        """
        Create a new reader object.

        :param path: A filesystem path to the wav file you wish to probe.

        :param info_encoding: The text encoding of the INFO metadata fields.
          latin_1/Win CP1252 has always been a pretty good guess for this.

        :param bext_encoding: The text encoding to use when decoding the string
          fields of the Broadcast-WAV extension. Per EBU 3285 this is ASCII
          but this parameter is available to you if you encounter a weirdo.
        """
        absolute_path = os.path.abspath(path)

        #: `file://` url for the file.
        self.url = pathlib.Path(absolute_path).as_uri()

        # for __repr__()
        self.path = absolute_path
        self.info_encoding = info_encoding
        self.bext_encoding = bext_encoding

        with open(path, 'rb') as f:
            chunks = parse_chunk(f)

            self.main_list = chunks.children
            f.seek(0)

            #: :class:`wavinfo.wave_reader.WavAudioFormat`
            self.fmt = self._get_format(f)

            #: :class:`wavinfo.wave_bext_reader.WavBextReader` with Broadcast-WAV metadata
            self.bext = self._get_bext(f, encoding=bext_encoding)

            #: :class:`wavinfo.wave_ixml_reader.WavIXMLFormat` with iXML metadata
            self.ixml = self._get_ixml(f)

            #: :class:`wavinfo.wave_info_reader.WavInfoChunkReader` with RIFF INFO metadata
            self.info = self._get_info(f, encoding=info_encoding)
            self.data = self._describe_data()

    def _find_chunk_data(self, ident, from_stream, default_none=False):
        top_chunks = (chunk for chunk in self.main_list if type(chunk) is ChunkDescriptor and chunk.ident == ident)
        chunk_descriptor = next(top_chunks, None) if default_none else next(top_chunks)
        return chunk_descriptor.read_data(from_stream) if chunk_descriptor else None

    def _describe_data(self):
        data_chunk = next(c for c in self.main_list if c.ident == b'data')

        return WavDataDescriptor(byte_count=data_chunk.length,
                                 frame_count=int(data_chunk.length / self.fmt.block_align))

    def _get_format(self, f):
        fmt_data = self._find_chunk_data(b'fmt ', f)

        # The format chunk is
        # audio_format    U16
        # channel_count   U16
        # sample_rate     U32   Note an integer
        # byte_rate       U32   == SampleRate * NumChannels * BitsPerSample/8
        # block_align     U16   == NumChannels * BitsPerSample/8
        # bits_per_sampl  U16
        packstring = "<HHIIHH"
        rest_starts = struct.calcsize(packstring)

        unpacked = struct.unpack(packstring, fmt_data[:rest_starts])

        # 0x0001	WAVE_FORMAT_PCM	PCM
        # 0x0003	WAVE_FORMAT_IEEE_FLOAT	IEEE float
        # 0x0006	WAVE_FORMAT_ALAW	8-bit ITU-T G.711 A-law
        # 0x0007	WAVE_FORMAT_MULAW	8-bit ITU-T G.711 µ-law
        # 0xFFFE	WAVE_FORMAT_EXTENSIBLE	Determined by SubFormat

        # https://sno.phy.queensu.ca/~phil/exiftool/TagNames/RIFF.html
        return WavAudioFormat(audio_format=unpacked[0],
                              channel_count=unpacked[1],
                              sample_rate=unpacked[2],
                              byte_rate=unpacked[3],
                              block_align=unpacked[4],
                              bits_per_sample=unpacked[5]
                              )

    def _get_info(self, f, encoding):
        finder = (chunk.signature for chunk in self.main_list if type(chunk) is ListChunkDescriptor)

        if b'INFO' in finder:
            return WavInfoChunkReader(f, encoding)

    def _get_bext(self, f, encoding):
        bext_data = self._find_chunk_data(b'bext', f, default_none=True)
        return WavBextReader(bext_data, encoding) if bext_data else None

    def _get_ixml(self, f):
        ixml_data = self._find_chunk_data(b'iXML', f, default_none=True)
        return None if ixml_data is None else WavIXMLFormat(ixml_data.rstrip(b'\0'))

    def walk(self):
        """
        Walk all of the available metadata fields.

        :yields: a string, the :scope: of the metadatum, the string :name: of the
        metadata field, and the value.
        """

        scopes = ('fmt', 'data')  # 'bext', 'ixml', 'info')

        for scope in scopes:
            attr = self.__getattribute__(scope)
            for field in attr._fields:
                yield scope, field, attr.__getattribute__(field)

            if self.bext is not None:
                bext_dict = (self.bext or {}).to_dict()
                for key in bext_dict.keys():
                    yield 'bext', key, bext_dict[key]

            if self.info is not None:
                info_dict = self.info.to_dict()
                for key in info_dict.keys():
                    yield 'info', key, info_dict[key]

    def __repr__(self):
        return 'WavInfoReader(%s, %s, %s)'.format(self.path, self.info_encoding, self.bext_encoding)



import struct
import binascii
from .umid_parser import UMIDParser


class WavBextReader:
    def __init__(self, bext_data, encoding):
        """
        Read Broadcast-WAV extended metadata.
        :param bext_data: The bytes-like data.
        :param encoding: The encoding to use when decoding the text fields of the
                 BEXT metadata scope. According to EBU Rec 3285 this shall be ASCII.
        """
        packstring = "<256s" + "32s" + "32s" + "10s" + "8s" + "QH" + "64s" + "hhhhh" + "180s"

        rest_starts = struct.calcsize(packstring)
        unpacked = struct.unpack(packstring, bext_data[:rest_starts])

        def sanitize_bytes(b):
            first_null = next((index for index, byte in enumerate(b) if byte == 0), None)
            trimmed = b if first_null is None else b[:first_null]
            decoded = trimmed.decode(encoding)
            return decoded

        #: Description. A free-text field up to 256 characters long.
        self.description = sanitize_bytes(unpacked[0])
        #: Originator. Usually the name of the encoding application, sometimes
        #: a artist name.
        self.originator = sanitize_bytes(unpacked[1])
        #: A unique identifier for the file, a serial number.
        self.originator_ref = sanitize_bytes(unpacked[2])
        #: Date of the recording, in the format YYY-MM-DD
        self.originator_date = sanitize_bytes(unpacked[3])
        #: Time of the recording, in the format HH:MM:SS.
        self.originator_time = sanitize_bytes(unpacked[4])
        #: The sample offset of the start of the file relative to an
        #: epoch, usually midnight the day of the recording. 
        self.time_reference = unpacked[5]
        #: A variable-length text field containing a list of processes and
        #: and conversions performed on the file.
        self.coding_history = sanitize_bytes(bext_data[rest_starts:])
        #: BEXT version. 
        self.version = unpacked[6]
        #: SMPTE 330M UMID of this audio file, 64 bytes are allocated though the UMID
        #: may only be 32 bytes long.
        self.umid = None
        #: EBU R128 Integrated loudness, in LUFS.
        self.loudness_value = None
        #: EBU R128 Loudness rante, in LUFS.
        self.loudness_range = None
        #: True peak level, in dBFS TP
        self.max_true_peak = None
        #: EBU R128 Maximum momentary loudness, in LUFS
        self.max_momentary_loudness = None
        #: EBU R128 Maximum short-term loudness, in LUFS.
        self.max_shortterm_loudness = None

        if self.version > 0:
            self.umid = unpacked[7]

        if self.version > 1:
            self.loudness_value = unpacked[8] / 100.0
            self.loudness_range = unpacked[9] / 100.0
            self.max_true_peak = unpacked[10] / 100.0
            self.max_momentary_loudness = unpacked[11] / 100.0
            self.max_shortterm_loudness = unpacked[12] / 100.0

    def to_dict(self):
        if self.umid is not None:
            umid_parsed = UMIDParser(self.umid)
            umid_str = umid_parsed.basic_umid_to_str()
        else:
            umid_str = None

        return {'description': self.description,
                'originator': self.originator,
                'originator_ref': self.originator_ref,
                'originator_date': self.originator_date,
                'originator_time': self.originator_time,
                'time_reference': self.time_reference,
                'version': self.version,
                'umid': umid_str,
                'coding_history': self.coding_history,
                'loudness_value': self.loudness_value,
                'loudness_range': self.loudness_range,
                'max_true_peak': self.max_true_peak,
                'max_momentary_loudness': self.max_momentary_loudness,
                'max_shortterm_loudness': self.max_shortterm_loudness
                }
